from datetime import timedelta
from decimal import Decimal

from rest_framework import serializers
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.metadata.configs.display.list_display import BaseTreeGroupLevelOption
from wbfdm.models import Instrument
from wbfdm.serializers import InvestableInstrumentRepresentationSerializer
from wbfdm.serializers.instruments.instruments import CompanyRepresentationSerializer, SecurityRepresentationSerializer

from wbportfolio.models import PortfolioRole, Trade, TradeProposal
from wbportfolio.models.transactions.claim import Claim
from wbportfolio.serializers.custodians import CustodianRepresentationSerializer
from wbportfolio.serializers.registers import RegisterRepresentationSerializer

from .transactions import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)


class CopyClaimRepresentationSerializer(wb_serializers.RepresentationSerializer):
    claimant = wb_serializers.StringRelatedField()
    product = wb_serializers.StringRelatedField()
    account = wb_serializers.StringRelatedField()
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:claim-detail")

    class Meta:
        model = Claim
        fields = (
            "id",
            "shares",
            "claimant",
            "bank",
            "product",
            "account",
            "_detail",
        )


class TradeProposalRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:tradeproposal-detail")

    class Meta:
        model = TradeProposal
        fields = ("id", "trade_date", "status", "_detail")


class TradeRepresentationSerializer(TransactionRepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:trade-detail")
    diff_shares = wb_serializers.DecimalField(max_digits=15, decimal_places=4)

    class Meta:
        model = Trade
        fields = ("claimed_shares", "diff_shares", "bank", "shares") + TransactionRepresentationSerializer.Meta.fields


class TradeClaimRepresentationSerializer(TradeRepresentationSerializer):
    def get_filter_params(self, request):
        claim_id = request.parser_context["view"].kwargs.get("pk", None)
        arg = {"marked_for_deletion": False}
        if claim_id:
            claim = Claim.objects.get(id=claim_id)
            arg.update(
                {
                    "underlying_instrument": claim.product.id,
                    "transaction_date__gte": claim.date - timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                    "transaction_date__lte": claim.date + timedelta(days=Trade.TRADE_WINDOW_INTERVAL),
                }
            )
            if claim.shares:
                if claim.shares > 0:
                    arg["shares__gte"] = claim.shares
                else:
                    arg["shares__lte"] = claim.shares
            return arg
        return {}


class TradeModelSerializer(TransactionModelSerializer):
    marked_for_deletion = wb_serializers.BooleanField(required=False, read_only=True)
    marked_as_internal = wb_serializers.BooleanField()

    approved_claimed_shares = wb_serializers.DecimalField(
        max_digits=10, decimal_places=2, required=False, read_only=True, default=Decimal(0.0)
    )
    pending_claimed_shares = wb_serializers.DecimalField(
        max_digits=10, decimal_places=2, required=False, read_only=True, default=Decimal(0.0)
    )
    completely_claimed = wb_serializers.BooleanField(required=False, read_only=True, default=Decimal(0.0))
    completely_claimed_if_approved = wb_serializers.BooleanField(required=False, read_only=True, default=Decimal(0.0))

    # We commented it because people needs to filter with lookup equals onto these fields
    # shares = wb_serializers.DecimalField(max_digits=16, decimal_places=2)
    # price = wb_serializers.DecimalField(max_digits=16, decimal_places=2)

    _register = RegisterRepresentationSerializer(source="register")
    claims = wb_serializers.PrimaryKeyRelatedField(queryset=Claim.objects.all(), label="Claims", many=True)
    _claims = CopyClaimRepresentationSerializer(source="claims", many=True)
    _custodian = CustodianRepresentationSerializer(source="custodian")
    _internal_trade = TradeRepresentationSerializer(
        source="internal_trade",
        filter_params={"only_internal_trade": True},
        optional_get_parameters={"underlying_instrument": "underlying_instrument", "transaction_date": "pivot_date"},
        depends_on=[{"field": "marked_as_internal", "options": {"activates_on": [True]}}],
    )

    @wb_serializers.register_only_instance_resource()
    def import_source_file_url(self, instance, request, user, *args, **kwargs):
        if PortfolioRole.is_portfolio_manager(user.profile, instance.underlying_instrument):
            if instance.import_source:
                return {"import_source": instance.import_source.file.url}
        return {}

    @wb_serializers.register_resource()
    def claims_list(self, instance, request, user):
        if instance.is_claimable:
            return {
                "claims": reverse(
                    "wbportfolio:trade-claim-list",
                    args=[instance.id],
                    request=request,
                )
            }
        return {}

    class Meta:
        model = Trade
        decorators = TransactionModelSerializer.Meta.decorators
        read_only_fields = (
            "id",
            "shares",
            "price",
            "transaction_subtype",
            "approved_claimed_shares",
            "pending_claimed_shares",
            "claimed_shares",
            "completely_claimed",
            "completely_claimed_if_approved",
            "register",
            "_register",
            "pending",
            "bank",
            "marked_for_deletion",
            "_claims",
            "claims",
            "transaction_type",
            "transaction_url_type",
            "transaction_underlying_type",
            "portfolio",
            "_portfolio",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_date",
            "book_date",
            "value_date",
            "currency",
            "_currency",
            "currency_fx_rate",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
            "external_id",
            "total_value_usd",
            "total_value_gross_usd",
            "_additional_resources",
        )

        fields = (
            "id",
            "shares",
            "price",
            "transaction_subtype",
            "approved_claimed_shares",
            "pending_claimed_shares",
            "claimed_shares",
            "completely_claimed",
            "completely_claimed_if_approved",
            "register",
            "_register",
            "pending",
            "bank",
            "marked_for_deletion",
            "_claims",
            "claims",
            "transaction_type",
            "transaction_url_type",
            "transaction_underlying_type",
            "portfolio",
            "_portfolio",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_date",
            "book_date",
            "value_date",
            "currency",
            "_currency",
            "currency_fx_rate",
            "total_value",
            "total_value_fx_portfolio",
            "total_value_gross",
            "total_value_gross_fx_portfolio",
            "external_id",
            "comment",
            "total_value_usd",
            "total_value_gross_usd",
            "_custodian",
            "custodian",
            "marked_as_internal",
            "internal_trade",
            "_internal_trade",
            "_additional_resources",
        )


class GetSecurityDefault:
    requires_context = True

    def __call__(self, serializer_instance):
        try:
            instance = serializer_instance.view.get_object()
            return instance.underlying_instrument.parent or instance.underlying_instrument
        except Exception:
            return None


class GetCompanyDefault:
    requires_context = True

    def __call__(self, serializer_instance):
        try:
            instance = serializer_instance.view.get_object()
            security = instance.underlying_instrument.parent or instance.underlying_instrument
            return security.parent or security
        except Exception:
            return None


class TradeTradeProposalModelSerializer(TradeModelSerializer):
    underlying_instrument_isin = wb_serializers.CharField(read_only=True)
    underlying_instrument_ticker = wb_serializers.CharField(read_only=True)
    underlying_instrument_refinitiv_identifier_code = wb_serializers.CharField(read_only=True)
    underlying_instrument_instrument_type = wb_serializers.CharField(read_only=True)

    company = wb_serializers.PrimaryKeyRelatedField(
        queryset=Instrument.objects.filter(level=0),
        required=False,
        read_only=lambda view: not view.new_mode,
        default=GetCompanyDefault(),
    )
    _company = CompanyRepresentationSerializer(source="company", required=False)

    security = wb_serializers.PrimaryKeyRelatedField(
        queryset=Instrument.objects.filter(is_security=True),
        required=False,
        read_only=lambda view: not view.new_mode,
        default=GetSecurityDefault(),
    )
    _security = SecurityRepresentationSerializer(
        source="security",
        optional_get_parameters={"company": "parent"},
        depends_on=[{"field": "company", "options": {}}],
        required=False,
    )
    underlying_instrument = wb_serializers.PrimaryKeyRelatedField(
        queryset=Instrument.objects.all(), label="Quote", read_only=lambda view: not view.new_mode
    )
    _underlying_instrument = InvestableInstrumentRepresentationSerializer(
        source="underlying_instrument",
        optional_get_parameters={"security": "parent"},
        depends_on=[{"field": "security", "options": {}}],
        tree_config=BaseTreeGroupLevelOption(clear_filter=True, filter_key="parent"),
    )

    status = wb_serializers.ChoiceField(default=Trade.Status.DRAFT, choices=Trade.Status.choices)
    target_weight = wb_serializers.DecimalField(max_digits=16, decimal_places=6, required=False, default=0)
    effective_weight = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=6, default=0)
    effective_shares = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=6, default=0)
    target_shares = wb_serializers.DecimalField(read_only=True, max_digits=16, decimal_places=6, default=0)

    def validate(self, data):
        data.pop("company", None)
        data.pop("security", None)
        if self.instance and "underlying_instrument" in data:
            raise serializers.ValidationError(
                {
                    "underlying_instrument": "You cannot modify the underlying instrument other than creating a new entry"
                }
            )
        effective_weight = self.instance._effective_weight if self.instance else Decimal(0.0)
        weighting = data.get("weighting", self.instance.weighting if self.instance else Decimal(0.0))
        if (target_weight := data.pop("target_weight", None)) is not None:
            weighting = target_weight - effective_weight
        if weighting >= 0:
            data["transaction_subtype"] = "BUY"
        else:
            data["transaction_subtype"] = "SELL"
        data["weighting"] = weighting
        return super().validate(data)

    class Meta:
        model = Trade
        percent_fields = ["effective_weight", "target_weight", "weighting"]
        read_only_fields = (
            "transaction_subtype",
            "shares",
            # "underlying_instrument",
            # "_underlying_instrument",
            "shares",
            "effective_shares",
            "target_shares",
        )
        fields = (
            "id",
            "shares",
            "underlying_instrument_isin",
            "underlying_instrument_ticker",
            "underlying_instrument_refinitiv_identifier_code",
            "underlying_instrument_instrument_type",
            "company",
            "_company",
            "security",
            "_security",
            "underlying_instrument",
            "_underlying_instrument",
            "transaction_subtype",
            "status",
            "comment",
            "effective_weight",
            "target_weight",
            "weighting",
            "trade_proposal",
            "order",
            "effective_shares",
            "target_shares",
        )


class ReadOnlyTradeTradeProposalModelSerializer(TradeTradeProposalModelSerializer):
    class Meta(TradeTradeProposalModelSerializer.Meta):
        read_only_fields = TradeTradeProposalModelSerializer.Meta.fields



from wbcore.metadata.configs.titles import TitleViewConfig
from wbfdm.models import Instrument


class AdjustmentTitleConfig(TitleViewConfig):
    def get_list_title(self):
        if instrument_id := self.view.kwargs.get("instrument_id", None):
            return f"Adjustments of {str(Instrument.objects.get(id=instrument_id))}"
        return "Adjustments"

    def get_instance_title(self):
        try:
            obj = self.view.get_object()
            return f"Adjustment: {str(obj)}"
        except AssertionError:
            return "Adjustment {{factor}} - {{date}} ({{_equity.computed_str}})"

    def get_create_title(self):
        return "New Adjustment"



import codecs
import csv
import datetime
import logging

from wbportfolio.import_export.utils import convert_string_to_number
from wbportfolio.models import Fees

logger = logging.getLogger("importers.parsers.sglux.fee")
# Shares class mapping between ticker to identifier
# {
#     "ATNRFRP": "UNIT/RG",
#     "ATNRFRC": "UNIT/RC",
#     "ATNRFRE": "UNIT/R1",
#     "ATNRFRU": "UNIT/R2",
#     "ATNRFPP": "UNIT/PG",
#     "ATNRFPC": "UNIT/PC",
#     "ATNRFPE": "UNIT/P1",
#     "ATNRFPU": "UNIT/P2",
#     "ATNRFZU": "UNIT/H",
#     "ATNRFAU": "UNIT/A",
#     "ATNRFBU": "UNIT/B",
#     "ATNRFFU": "UNIT/F"
# }


def parse(import_source):
    fee_file = import_source.file.open()
    fee_file = codecs.iterdecode(fee_file, "latin1")

    # Read file into a CSV Dict Reader
    fee_reader = csv.DictReader(fee_file, delimiter=",")

    # Iterate through the CSV File and parse the data into a list
    data = list()

    for fee_data in fee_reader:
        if (fee_description := fee_data["Fees description"]) and fee_description.startswith("Investment Manager fees"):
            share_class = fee_data["Class"]  # .split("\\")[1][0]
            date = datetime.datetime.strptime(fee_data["NAV Date"], "%Y/%m/%d")
            data.append(
                {
                    "linked_product": {
                        "parent__identifier": fee_data["Code"],
                        "currency__key": fee_data["Local ccy"],
                        "identifier": share_class,
                    },
                    "transaction_date": date.strftime("%Y-%m-%d"),
                    "calculated": False,
                    "transaction_subtype": Fees.Type.MANAGEMENT,
                    "total_value": round(convert_string_to_number(fee_data.get("Fees - Local ccy", 0)), 6),
                }
            )

    fee_file.close()
    return {"data": data}



from decimal import Decimal

from django.db import models

from wbportfolio.import_export.handlers.dividend import DividendImportHandler

from .transactions import ShareMixin, Transaction


class DividendTransaction(Transaction, ShareMixin, models.Model):
    import_export_handler_class = DividendImportHandler
    retrocession = models.FloatField(default=1)

    def save(self, *args, **kwargs):
        super().save(*args, factor=Decimal(self.retrocession), **kwargs)



from django.contrib import admin
from django.db.models import Prefetch

from wbportfolio.admin.transactions import TransactionModelAdmin
from wbportfolio.models import Fees, Portfolio
from wbportfolio.models.transactions.fees import FeeCalculation


@admin.register(Fees)
class FeesAdmin(TransactionModelAdmin):
    list_filter = [*TransactionModelAdmin.list_filter, "linked_product"]
    list_display = ["transaction_subtype", *TransactionModelAdmin.list_display[1:], "calculated"]
    fieldsets = (
        ("Transaction Information", TransactionModelAdmin.fieldsets[0][1]),
        ("Fees Information", {"fields": (("transaction_subtype", "calculated", "linked_product"),)}),
    )
    autocomplete_fields = [*TransactionModelAdmin.autocomplete_fields, "linked_product"]

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .prefetch_related(Prefetch("portfolio", queryset=Portfolio.objects.all().only("id", "name")))
        )

    def calculate(self, request, queryset):
        for fees in queryset:
            fees.calculate_as_task.delay(fees)

    calculate.short_description = "Recalculate selected Fees"
    actions = [calculate]


@admin.register(FeeCalculation)
class FeeCalculationModelAdmin(admin.ModelAdmin):
    search_fields = ("name",)



import datetime
import re

import xlrd


def file_name_parse(file_name):
    isin = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", file_name)

    assert len(isin) == 1, "Not exactly 1 isin found in the filename"

    return {"isin": isin[0]}


def parse(import_source):
    book = xlrd.open_workbook(file_contents=import_source.file.read())

    parts = file_name_parse(import_source.file.name)
    isin = parts["isin"]

    equity_sheet = book.sheet_by_name("Equity")
    cash_sheet = book.sheet_by_name("Cash")

    date = None
    first_row = None
    last_row = None

    for row_index, row in enumerate(equity_sheet.get_rows()):
        if "Statement as at Close of Business" in row[1].value:
            date = datetime.datetime.strptime(row[1].value.split(":")[1].strip(), "%d-%b-%Y").date()

        if "Instrument" in row[1].value:
            first_row = row_index + 1

        if row[1].value == "" and first_row is not None:
            last_row = row_index
            break

    data = list()
    for row in range(first_row, last_row, 1):
        initial_shares = round(equity_sheet.cell_value(row, 8), 4)
        close = round(equity_sheet.cell_value(row, 9), 4)
        initial_currency_fx_rate = round(equity_sheet.cell_value(row, 11), 4)

        bbg = equity_sheet.cell_value(row, 3)
        ric = equity_sheet.cell_value(row, 2)
        try:
            ticker, exchange = bbg.split(" ")[0:2]
        except Exception:
            ticker, exchange = bbg, None
        exchange_data = {"bbg_exchange_codes": exchange}
        data.append(
            {
                "underlying_quote": {
                    "instrument_type": "equity",
                    "ticker": ticker,
                    "exchange": exchange_data,
                    "name": equity_sheet.cell_value(row, 1),
                    "currency__key": equity_sheet.cell_value(row, 10),
                    "refinitiv_identifier_code": ric,
                },
                "portfolio": {"instrument_type": "product", "isin": isin},
                "is_estimated": False,
                "exchange": exchange_data,
                "initial_shares": initial_shares,
                "date": date.strftime("%Y-%m-%d"),
                "asset_valuation_date": date.strftime("%Y-%m-%d"),
                "initial_price": close,
                "currency__key": equity_sheet.cell_value(row, 10),
                "initial_currency_fx_rate": initial_currency_fx_rate,
            }
        )

    for row_index, row in enumerate(cash_sheet.get_rows()):
        if "Currency" == row[1].value:
            data.append(
                {
                    "underlying_quote": {
                        "instrument_type": "cash",
                        "currency__key": cash_sheet.cell_value(row_index + 1, 1),
                    },
                    "portfolio": {"instrument_type": "product", "isin": isin},
                    "is_estimated": False,
                    "name": cash_sheet.cell_value(row_index + 1, 1),
                    "initial_shares": round(cash_sheet.cell_value(row_index + 1, 2), 4),
                    "date": date.strftime("%Y-%m-%d"),
                    "asset_valuation_date": date.strftime("%Y-%m-%d"),
                    "initial_price": 1.0,
                    "currency__key": cash_sheet.cell_value(row_index + 1, 1),
                    "initial_currency_fx_rate": 1.0,
                }
            )
            break

    return {"data": data}



from typing import Generator

import pandas as pd
from wbfdm.analysis.esg.enums import ESGAggregation
from wbfdm.analysis.esg.esg_analysis import DataLoader, get_esg_df
from wbfdm.contrib.metric.decorators import register
from wbfdm.contrib.metric.dto import Metric, MetricField, MetricKey

from wbportfolio.models import Portfolio

from .base import PortfolioMetricBaseBackend
from .portfolio_base import DataLoader as PortfolioDataLoader


def _generate_metric_key_from_enum():
    metric_keys = []
    for key, label in ESGAggregation.choices():
        metric_keys.append(
            MetricKey(
                key=f"portfolio_esg_{key.lower()}",
                label=label,
                subfields=[
                    MetricField(key="score", label="Score"),
                ],
            )
        )
    return metric_keys


PORTFOLIO_ESG_KEYS = _generate_metric_key_from_enum()


@register()
class PortfolioESGMetricBackend(PortfolioMetricBaseBackend):
    keys = PORTFOLIO_ESG_KEYS

    def _get_metrics_for_esg_aggregation(
        self, instruments, weights: pd.Series, total_value_fx_usd: pd.Series, esg_aggregation: ESGAggregation
    ) -> pd.Series:
        dl = DataLoader(
            weights, get_esg_df(instruments, esg_aggregation.get_esg_code()), self.val_date, total_value_fx_usd
        )
        df = dl.compute(esg_aggregation)

        return df.rename("score").to_frame()

    def compute_metrics(self, basket: Portfolio) -> Generator[Metric, None, None]:
        val_date = self._get_valid_date(basket)
        portfolio_dataloader = PortfolioDataLoader(basket, val_date)
        base_df = portfolio_dataloader.base_df

        if not base_df.empty:
            for esg in ESGAggregation:
                df = self._get_metrics_for_esg_aggregation(
                    portfolio_dataloader.instruments, base_df["weighting"], base_df["total_value_fx_usd"], esg
                )
                df_aggregated = df.sum(axis=0)
                metrics = self.convert_df_into_metrics(df, f"portfolio_esg_{esg.name.lower()}", basket.id, val_date)
                aggregated_metrics = self.convert_df_into_metrics(
                    df_aggregated,
                    f"portfolio_esg_{esg.name.lower()}",
                    basket.id,
                    val_date,
                    dependency_metrics=metrics,
                )
                yield from metrics
                yield from aggregated_metrics



from contextlib import suppress

import pandas as pd

from wbportfolio.models import Product, Trade


def parse(import_source):
    data = list()
    df = pd.DataFrame()
    try:
        df = pd.read_excel(
            import_source.file.read(),
            engine="openpyxl",
            index_col=1,
            sheet_name="TRANSACTION CONSOLIDATION",
            skiprows=3,
        ).dropna(axis=1)
        df["DATE"] = pd.to_datetime(df["DATE"], dayfirst=True)
        df = df[df["EVENT"] == "AMC Cash"]
    except (ValueError, IndexError):
        pass
    if not df.empty:
        for external_id, trade in df.to_dict("index").items():
            with suppress(Product.DoesNotExist):
                product = Product.objects.get(isin=trade["AMC ISIN"])
                shares = (-1 * trade["TOTAL QUANTITY"]) / product.share_price
                price = trade["GROSS PRICE"] * product.share_price
                portfolio = product.primary_portfolio
                data.append(
                    {
                        "underlying_instrument": {"id": product.id, "instrument_type": "product"},
                        "portfolio": portfolio.id,
                        "transaction_date": trade["DATE"].strftime("%Y-%m-%d"),
                        "shares": shares,
                        "external_id": external_id,
                        "transaction_subtype": Trade.Type.REDEMPTION if shares < 0 else Trade.Type.SUBSCRIPTION,
                        "bank": "Leonteq Cash Transfer",
                        "currency__key": product.currency.key,
                        "price": price,
                        "total_value": trade["NET AMOUNT"],
                        "total_value_gross": trade["GROSS AMOUNT"],
                        "currency_fx_rate": trade["BOOKING FX"],
                        "total_value_fx_portfolio": trade["BOOKING NET AMOUNT"],
                    }
                )

    return {"data": data}



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class PortfolioInstrumentPreferredClassificationThroughDisplayConfig(DisplayViewConfig):
    def get_instance_display(self) -> Display:
        return create_simple_display([["instrument", "classification"]])

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="instrument", label="Instrument"),
                dp.Field(key="classification_group", label="Preferred Classification Group"),
                dp.Field(key="classification", label="Preferred Classification"),
            ]
        )


class InstrumentPortfolioThroughPortfolioModelDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="instrument", label="Instrument"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["instrument"],
            ]
        )



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class FeeEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None


class FeesPortfolioEndpointConfig(FeeEndpointConfig):
    pass


class FeesAggregatedPortfolioPandasEndpointConfig(EndpointViewConfig):
    pass



from wbcore.metadata.configs.endpoints import EndpointViewConfig


class TransactionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        model = "{{transaction_url_type}}"
        return f"{self.request.scheme}://{self.request.get_host()}/api/portfolio/{model}/"


class TransactionPortfolioEndpointConfig(TransactionEndpointConfig):
    pass



import pandas as pd

from wbportfolio.models import Fees

from .utils import get_perf_fee_isin

FIELD_MAP = {
    "LastPriceDate": "transaction_date",
    "EndOfDay": "value_date",
    "AccrValueInstrCcy": "total_value",
    "FxRate": "currency_fx_rate",
    "TradingCurrency": "currency__key",
    "#ClientName": "linked_product",
    "InstrumentName": "comment",
}

PERF_FEES_INSTRUMENT_ISIN = "CH0040602242"


def parse(import_source):
    perf_fee_isin = get_perf_fee_isin(import_source.source)
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=";")
    df = df.loc[df["ISIN"] == perf_fee_isin, :]
    data = []
    if not df.empty:
        df = df.rename(columns=FIELD_MAP)
        df = df.drop(columns=df.columns.difference(FIELD_MAP.values()))
        df.total_value = -df.total_value
        df["transaction_subtype"] = Fees.Type.MANAGEMENT
        df["transaction_date"] = pd.to_datetime(df["transaction_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["value_date"] = pd.to_datetime(df["value_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df = df.dropna(subset=["linked_product"])
        df.linked_product = df.linked_product.apply(lambda x: {"identifier": x})
        data = df.to_dict("records")
    return {"data": data}



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig

from wbportfolio.models import TradeProposal


class TradeProposalPortfolioEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbportfolio:portfolio-tradeproposal-list", args=[self.view.kwargs["portfolio_id"]], request=self.request
        )

    def get_delete_endpoint(self, **kwargs):
        if trade_proposal_id := self.view.kwargs.get("pk", None):
            trade_proposal = TradeProposal.objects.get(id=trade_proposal_id)
            if trade_proposal.status == TradeProposal.Status.DRAFT:
                return reverse("wbportfolio:tradeproposal-list", args=[], request=self.request)
        return None



from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal

from celery import shared_task
from django.contrib import admin
from django.db import models
from django.db.models import (
    Case,
    DateField,
    ExpressionWrapper,
    F,
    OuterRef,
    Q,
    Subquery,
    Sum,
    When,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from django_fsm import GET_STATE, FSMField, transition
from ordered_model.models import OrderedModel, OrderedModelManager, OrderedModelQuerySet
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.models import WBModel
from wbcore.signals.models import pre_collection
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

from wbportfolio.import_export.handlers.trade import TradeImportHandler
from wbportfolio.models.asset import AssetPosition
from wbportfolio.models.custodians import Custodian
from wbportfolio.models.roles import PortfolioRole
from wbportfolio.pms.typing import Trade as TradeDTO

from .transactions import ShareMixin, Transaction


class TradeQueryset(OrderedModelQuerySet):
    def annotate_base_info(self):
        return self.annotate(
            last_effective_date=Subquery(
                AssetPosition.unannotated_objects.filter(
                    date__lte=OuterRef("value_date"),
                    portfolio=OuterRef("portfolio"),
                )
                .order_by("-date")
                .values("date")[:1]
            ),
            effective_weight=Coalesce(
                Subquery(
                    AssetPosition.unannotated_objects.filter(
                        underlying_quote=OuterRef("underlying_instrument"),
                        date=OuterRef("last_effective_date"),
                        portfolio=OuterRef("portfolio"),
                    )
                    .values("portfolio")
                    .annotate(s=Sum("weighting"))
                    .values("s")[:1]
                ),
                Decimal(0),
            ),
            target_weight=F("effective_weight") + F("weighting"),
            effective_shares=Coalesce(
                Subquery(
                    AssetPosition.objects.filter(
                        underlying_quote=OuterRef("underlying_instrument"),
                        date=OuterRef("last_effective_date"),
                        portfolio=OuterRef("portfolio"),
                    )
                    .values("portfolio")
                    .annotate(s=Sum("shares"))
                    .values("s")[:1]
                ),
                Decimal(0),
            ),
            target_shares=F("effective_shares") + F("shares"),
        )


class DefaultTradeManager(OrderedModelManager):
    """This manager is expect to be the trade default manager and annotate by default the effective weight (extracted
    from the associated portfolio) and the target weight as an addition between the effective weight and the delta weight
    """

    def __init__(self, with_annotation: bool = False, *args, **kwargs):
        self.with_annotation = with_annotation
        super().__init__(*args, **kwargs)

    def get_queryset(self) -> TradeQueryset:
        qs = TradeQueryset(self.model, using=self._db)
        if self.with_annotation:
            qs = qs.annotate_base_info()
        return qs


class ValidCustomerTradeManager(DefaultTradeManager):
    def __init__(self, without_internal_trade: bool = False):
        self.without_internal_trade = without_internal_trade
        super().__init__()

    def get_queryset(self):
        qs = (
            super()
            .get_queryset()
            .filter(
                transaction_subtype__in=[Trade.Type.SUBSCRIPTION, Trade.Type.REDEMPTION],
                marked_for_deletion=False,
                pending=False,
            )
        )
        if self.without_internal_trade:
            qs = qs.exclude(marked_as_internal=True)
        return qs


class Trade(ShareMixin, Transaction, OrderedModel, WBModel):
    import_export_handler_class = TradeImportHandler

    TRADE_WINDOW_INTERVAL = 7

    class Status(models.TextChoices):
        DRAFT = "DRAFT", "Draft"
        SUBMIT = "SUBMIT", "Submit"
        EXECUTED = "EXECUTED", "Executed"
        CONFIRMED = "CONFIRMED", "Confirmed"
        FAILED = "FAILED", "Failed"

    class Type(models.TextChoices):
        REBALANCE = "REBALANCE", "Rebalance"
        DECREASE = "DECREASE", "Decrease"
        INCREASE = "INCREASE", "Increase"
        SUBSCRIPTION = "SUBSCRIPTION", "Subscription"
        REDEMPTION = "REDEMPTION", "Redemption"
        BUY = "BUY", "Buy"
        SELL = "SELL", "Sell"
        NO_CHANGE = "NO_CHANGE", "No Change"  # default transaction subtype if weighing is 0

    external_identifier2 = models.CharField(
        max_length=255,
        null=True,
        blank=True,
        help_text="A second external identifier that was supplied.",
        verbose_name="External Identifier 2",
    )

    transaction_subtype = models.CharField(
        max_length=32, default=Type.BUY, choices=Type.choices, verbose_name="Trade Type"
    )
    status = FSMField(default=Status.CONFIRMED, choices=Status.choices, verbose_name="Status")
    weighting = models.DecimalField(
        max_digits=16,
        decimal_places=6,
        default=Decimal(0),
        help_text="The weight to be multiplied against the target",
        verbose_name="Weight",
    )
    bank = models.CharField(
        max_length=255,
        help_text="The bank/counterparty/custodian the trade went through.",
        verbose_name="Counterparty",
    )
    custodian = models.ForeignKey(
        "wbportfolio.Custodian", null=True, blank=True, on_delete=models.SET_NULL, related_name="trades"
    )
    marked_for_deletion = models.BooleanField(
        default=False,
        help_text="If this is checked, then the trade is supposed to be deleted.",
        verbose_name="To be deleted",
    )

    # Only valid for subscription and redemption trade
    marked_as_internal = models.BooleanField(
        default=False,
        help_text="If this is checked, then this subscription or redemption is considered internal and will not be considered in any AUM computation",
        verbose_name="Internal",
    )
    internal_trade = models.OneToOneField(
        "wbportfolio.Trade",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="internal_subscription_redemption_trade",
    )

    pending = models.BooleanField(default=False)
    exclude_from_history = models.BooleanField(default=False)
    register = models.ForeignKey(
        to="wbportfolio.Register",
        null=True,
        blank=True,
        related_name="trades",
        on_delete=models.PROTECT,
    )

    trade_proposal = models.ForeignKey(
        to="wbportfolio.TradeProposal",
        null=True,
        blank=True,
        related_name="trades",
        on_delete=models.CASCADE,
        help_text="The Trade Proposal this trade is coming from",
    )
    claimed_shares = models.DecimalField(
        max_digits=15,
        decimal_places=4,
        default=Decimal(0),
        help_text="The number of shares that were claimed.",
        verbose_name="Claimed Shares",
    )
    diff_shares = models.GeneratedField(
        expression=F("shares") - F("claimed_shares"),
        output_field=models.DecimalField(max_digits=15, decimal_places=4),
        db_persist=True,
    )
    objects = DefaultTradeManager()
    annotated_objects = DefaultTradeManager(with_annotation=True)
    valid_customer_trade_objects = ValidCustomerTradeManager()
    valid_external_customer_trade_objects = ValidCustomerTradeManager(without_internal_trade=True)

    @transition(
        field=status,
        source=Status.DRAFT,
        target=GET_STATE(
            lambda self, **kwargs: (
                self.Status.SUBMIT if self.last_underlying_quote_price is not None else self.Status.FAILED
            ),
            states=[Status.SUBMIT, Status.FAILED],
        ),
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:trade",),
                icon=WBIcon.SEND.icon,
                key="submit",
                label="Submit",
                action_label="Submit",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
        on_error="FAILED",
    )
    def submit(self, by=None, description=None, **kwargs):
        if not self.last_underlying_quote_price:
            self.comment = f"Trade failed because no price is found for {self.underlying_instrument.computed_str} on {self.transaction_date:%Y-%m-%d}"

    def can_submit(self):
        pass

    @transition(
        field=status,
        source=Status.DRAFT,
        target=Status.FAILED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
    )
    def fail(self, **kwargs):
        self.comment = f"Trade failed because no price is found for {self.underlying_instrument.computed_str} on {self.transaction_date:%Y-%m-%d}"

    @cached_property
    def last_underlying_quote_price(self) -> InstrumentPrice | None:
        try:
            # we try t0 first
            return InstrumentPrice.objects.filter_only_valid_prices().get(
                instrument=self.underlying_instrument, date=self.transaction_date
            )
        except InstrumentPrice.DoesNotExist:
            with suppress(InstrumentPrice.DoesNotExist):
                # we fall back to the latest price before t0
                return (
                    InstrumentPrice.objects.filter_only_valid_prices()
                    .filter(instrument=self.underlying_instrument, date__lte=self.value_date)
                    .latest("date")
                )

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.EXECUTED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:trade",),
                icon=WBIcon.CONFIRM.icon,
                key="execute",
                label="Execute",
                action_label="Execute",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def execute(self, **kwargs):
        with suppress(ValueError):
            asset = self.get_asset()
            AssetPosition.unannotated_objects.update_or_create(
                underlying_quote=asset.underlying_quote,
                portfolio_created=asset.portfolio_created,
                portfolio=asset.portfolio,
                date=asset.date,
                defaults={
                    "initial_currency_fx_rate": asset.initial_currency_fx_rate,
                    "initial_price": asset.initial_price,
                    "initial_shares": asset.initial_shares,
                    "underlying_quote_price": asset.underlying_quote_price,
                    "asset_valuation_date": asset.asset_valuation_date,
                    "currency": asset.currency,
                    "is_estimated": asset.is_estimated,
                    "weighting": asset.weighting,
                },
            )

    def can_execute(self):
        if not self.last_underlying_quote_price:
            return {"underlying_instrument": [_("Cannot execute a trade without a valid quote price")]}
        if not self.portfolio.is_manageable:
            return {
                "portfolio": [_("The portfolio needs to be a model portfolio in order to execute this trade manually")]
            }

    @transition(
        field=status,
        source=Status.EXECUTED,
        target=Status.CONFIRMED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:trade",),
                icon=WBIcon.CONFIRM.icon,
                key="confirm",
                label="Confirm",
                action_label="Confirme",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def confirm(self, by=None, description=None, **kwargs):
        pass

    def can_confirm(self):
        pass

    @transition(
        field=status,
        source=Status.SUBMIT,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:trade",),
                icon=WBIcon.UNDO.icon,
                key="backtodraft",
                label="Back to Draft",
                action_label="backtodraft",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def backtodraft(self, **kwargs):
        pass

    @transition(
        field=status,
        source=Status.EXECUTED,
        target=Status.DRAFT,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(
            user.profile, portfolio=instance.portfolio
        ),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:trade",),
                icon=WBIcon.UNDO.icon,
                key="reverte",
                label="Revert",
                action_label="revert",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def revert(self, to_date=None, **kwargs):
        with suppress(AssetPosition.DoesNotExist):
            asset = AssetPosition.unannotated_objects.get(
                underlying_quote=self.underlying_instrument,
                portfolio=self.portfolio,
                date=self.transaction_date,
                is_estimated=False,
            )
            asset.set_weighting(asset.weighting - self.weighting)
            asset.save()

    @property
    def product(self):
        from wbportfolio.models.products import Product

        try:
            return Product.objects.get(id=self.underlying_instrument.id)
        except Product.DoesNotExist:
            return None

    @property
    @admin.display(description="Last Effective Date")
    def _last_effective_date(self) -> date:
        if hasattr(self, "last_effective_date"):
            return self.last_effective_date
        elif (
            assets := AssetPosition.unannotated_objects.filter(
                underlying_quote=self.underlying_instrument,
                date__lte=self.value_date,
                portfolio=self.portfolio,
            )
        ).exists():
            return assets.latest("date").date

    @property
    @admin.display(description="Effective Weight")
    def _effective_weight(self) -> Decimal:
        return getattr(
            self,
            "effective_weight",
            AssetPosition.unannotated_objects.filter(
                underlying_quote=self.underlying_instrument,
                date=self._last_effective_date,
                portfolio=self.portfolio,
            ).aggregate(s=Sum("weighting"))["s"]
            or Decimal(0),
        )

    @property
    @admin.display(description="Effective Shares")
    def _effective_shares(self) -> Decimal:
        return getattr(
            self,
            "effective_shares",
            AssetPosition.objects.filter(
                underlying_quote=self.underlying_instrument,
                date=self.transaction_date,
                portfolio=self.portfolio,
            ).aggregate(s=Sum("shares"))["s"]
            or Decimal(0),
        )

    @property
    @admin.display(description="Target Weight")
    def _target_weight(self) -> Decimal:
        return getattr(self, "target_weight", self._effective_weight + self.weighting)

    @property
    @admin.display(description="Target Shares")
    def _target_shares(self) -> Decimal:
        return getattr(self, "target_shares", self._effective_shares + self.shares)

    order_with_respect_to = "trade_proposal"

    class Meta(OrderedModel.Meta):
        verbose_name = "Trade"
        verbose_name_plural = "Trades"
        constraints = [
            models.CheckConstraint(
                check=models.Q(marked_as_internal=False)
                | (
                    models.Q(marked_as_internal=True)
                    & models.Q(transaction_subtype__in=["REDEMPTION", "SUBSCRIPTION"])
                ),
                name="marked_as_internal_only_for_subred",
            ),
            models.CheckConstraint(
                check=models.Q(internal_trade__isnull=True)
                | (
                    models.Q(internal_trade__isnull=False)
                    & models.Q(transaction_subtype__in=["REDEMPTION", "SUBSCRIPTION"])
                ),
                name="internal_trade_set_only_for_subred",
            ),
        ]
        # notification_email_template = "portfolio/email/trade_notification.html"

    def __init__(self, *args, target_weight: Decimal | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        if target_weight is not None:  # if target weight is provided, we guess the corresponding weighting
            self.weighting = Decimal(target_weight) - self._effective_weight
            self._set_transaction_subtype()

    def save(self, *args, **kwargs):
        if self.trade_proposal:
            self.portfolio = self.trade_proposal.portfolio
            self.transaction_date = self.trade_proposal.trade_date
            self.value_date = self.trade_proposal.last_effective_date
            if not self.portfolio.only_weighting:
                with suppress(ValueError):
                    self.shares = self.trade_proposal.get_estimated_shares(self.weighting, self.underlying_instrument)

        if not self.custodian and self.bank:
            self.custodian = Custodian.get_by_mapping(self.bank)
        if self.price is None:
            # we try to get the price if not provided directly from the underlying instrument
            with suppress(Exception):
                self.price = self.underlying_instrument.get_price(self.value_date)

        self.transaction_type = Transaction.Type.TRADE

        if self.transaction_subtype is None or self.trade_proposal:
            # if subtype not provided, we extract it automatically from the existing data.
            self._set_transaction_subtype()
        if self.id and hasattr(self, "claims"):
            self.claimed_shares = self.trade.claims.filter(status="APPROVED").aggregate(s=Sum("shares"))[
                "s"
            ] or Decimal(0)
        if self.internal_trade:
            self.marked_as_internal = True
        super().save(*args, **kwargs)

    def _set_transaction_subtype(self):
        if self.weighting == 0:
            self.transaction_subtype = Trade.Type.NO_CHANGE
        if self.underlying_instrument.instrument_type.key == "product":
            if self.shares is not None:
                if self.shares > 0:
                    self.transaction_subtype = Trade.Type.SUBSCRIPTION
                elif self.shares < 0:
                    self.transaction_subtype = Trade.Type.REDEMPTION
        elif self.weighting is not None:
            if self.weighting > 0:
                if self._effective_weight:
                    self.transaction_subtype = Trade.Type.INCREASE
                else:
                    self.transaction_subtype = Trade.Type.BUY
            elif self.weighting < 0:
                if self._target_weight:
                    self.transaction_subtype = Trade.Type.DECREASE
                else:
                    self.transaction_subtype = Trade.Type.SELL
        else:
            self.transaction_subtype = Trade.Type.REBALANCE

    def get_transaction_subtype(self) -> str:
        """
        Return the expected transaction subtype based n

        """

    def get_asset(self) -> AssetPosition:
        last_underlying_quote_price = self.last_underlying_quote_price
        if not last_underlying_quote_price:
            raise ValueError("No price found")
        asset = AssetPosition(
            underlying_quote=self.underlying_instrument,
            portfolio_created=None,
            portfolio=self.portfolio,
            date=self.transaction_date,
            initial_currency_fx_rate=self.currency_fx_rate,
            weighting=self._target_weight,
            initial_price=self.last_underlying_quote_price.net_value,
            initial_shares=None,
            underlying_quote_price=self.last_underlying_quote_price,
            asset_valuation_date=self.transaction_date,
            currency=self.currency,
            is_estimated=False,
        )
        asset.set_weighting(self._target_weight)
        asset.pre_save()
        return asset

    def delete(self, **kwargs):
        pre_collection.send(sender=self.__class__, instance=self)
        super().delete(**kwargs)

    def __str__(self):
        ticker = f"{self.underlying_instrument.ticker}:" if self.underlying_instrument.ticker else ""
        return f"{ticker}{self.shares} ({self.bank})"

    def _build_dto(self) -> TradeDTO:
        """
        Data Transfer Object
        Returns:
            DTO trade object
        """
        return TradeDTO(
            id=self.id,
            underlying_instrument=self.underlying_instrument.id,
            effective_weight=self._effective_weight,
            target_weight=self._target_weight,
            instrument_type=self.underlying_instrument.security_instrument_type.id,
            currency=self.underlying_instrument.currency,
            date=self.transaction_date,
        )

    def get_alternative_valid_trades(self, share_delta: float = 0):
        return Trade.objects.filter(
            Q(underlying_instrument=self.underlying_instrument)
            & Q(portfolio=self.portfolio)
            & (
                Q(transaction_date__gte=self.transaction_date - timedelta(days=self.TRADE_WINDOW_INTERVAL))
                & Q(transaction_date__lte=self.transaction_date + timedelta(days=self.TRADE_WINDOW_INTERVAL))
            )
            & Q(transaction_subtype=self.transaction_subtype)
            & Q(shares__gte=self.shares * Decimal(1 - share_delta))
            & Q(shares__lte=self.shares * Decimal(1 + share_delta))
            & Q(marked_for_deletion=False)
            & Q(claims__isnull=True)
            & Q(pending=False)
        ).exclude(id=self.id)

    @property
    def is_claimable(self) -> bool:
        return self.is_customer_trade and not self.marked_for_deletion and not self.pending

    @property
    def is_customer_trade(self) -> bool:
        return self.transaction_subtype in [Trade.Type.REDEMPTION.name, Trade.Type.SUBSCRIPTION.name]

    @classmethod
    def subquery_shares_per_underlying_instrument(
        cls, val_date, underlying_instrument_name="pk", only_customer_trade=True
    ):
        """Returns a Subquery that returns the shares at a certain price date
        or 0

        Arguments:
            val_date {datetime.date} -- The  date that is used to determine which tradesare filtered

        Keyword Arguments:
            underlying_instrument_name {str} -- The reference to the underlying_instrument pk of the outer query (default: {"pk"})

        Returns:
            django.db.models.Subquery -- Subquery containing the sum of shares of each underlying_instrument
        """

        qs = cls.valid_customer_trade_objects
        if not only_customer_trade:
            qs = cls.objects
        qs = qs.filter(
            underlying_instrument=OuterRef(underlying_instrument_name),
            transaction_date__lt=val_date,
        )
        return Coalesce(
            Subquery(
                qs.values("underlying_instrument").annotate(sum_shares=Sum("shares")).values("sum_shares")[:1],
                output_field=models.DecimalField(),
            ),
            Decimal(0),
        )

    def link_to_internal_trade(self):
        qs = Trade.objects.filter(
            Q(underlying_instrument__instrument_type__key="product")
            & Q(shares=self.shares)
            & Q(underlying_instrument=self.underlying_instrument)
            & Q(transaction_date__gte=self.transaction_date - timedelta(days=self.TRADE_WINDOW_INTERVAL))
            & Q(transaction_date__lte=self.transaction_date + timedelta(days=self.TRADE_WINDOW_INTERVAL))
        ).exclude(id=self.id)
        if self.transaction_subtype in [Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION]:
            qs = qs.exclude(transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION])
            if qs.count() == 1:
                self.internal_trade = qs.first()
                self.save()
        else:
            qs = qs.filter(transaction_subtype__in=[Trade.Type.REDEMPTION, Trade.Type.SUBSCRIPTION])
            if qs.count() == 1:
                trade = qs.first()
                trade.internal_trade = self
                trade.save()

    @classmethod
    def subquery_net_money(
        cls, date_gte=None, date_lte=None, underlying_instrument_name="pk", only_positive=False, only_negative=False
    ):
        """Return a subquery which computes the net negative/positive money per underlying_instrument

        Arguments:
            val_date1 {datetime.date} -- The start date, including
            val_date2 {datetime.date} -- The end date, including

        Keyword Arguments:
            underlying_instrument_name {str} -- The reference to the underlying_instrument pk from the outer query (default: {"pk"})

        Returns:
            django.db.models.Subquery -- The subquery containing the net negative money per underlying_instrument
        """
        qs = cls.valid_external_customer_trade_objects.annotate(
            date_considered=ExpressionWrapper(F("transaction_date") + 1, output_field=DateField())
        )

        if date_gte:
            qs = qs.filter(date_considered__gte=date_gte)
        if date_lte:
            qs = qs.filter(date_considered__lte=date_lte)

        if only_positive:
            qs = qs.filter(shares__gt=0)
        elif only_negative:
            qs = qs.filter(shares__lt=0)
        return Coalesce(
            Subquery(
                qs.filter(underlying_instrument=OuterRef(underlying_instrument_name))
                .annotate(
                    _price=Case(
                        When(
                            price__isnull=True,
                            then=InstrumentPrice.subquery_closest_value(
                                "net_value",
                                date_name="date_considered",
                                instrument_pk_name="underlying_instrument__pk",
                            ),
                        ),
                        default=F("price"),
                    ),
                    net_value=ExpressionWrapper(F("shares") * F("_price"), output_field=models.FloatField()),
                )
                .values("underlying_instrument")
                .annotate(sum_net_value=Sum(F("net_value")))
                .values("sum_net_value"),
                output_field=models.FloatField(),
            ),
            0.0,
        )

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:trade"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:traderepresentation-list"

    @classmethod
    def get_representation_label_key(cls):
        return "{{|:-}}{{transaction_date}}{{|::}}{{bank}}{{|-:}} {{claimed_shares}} / {{shares}} (∆ {{diff_shares}})"


@shared_task
def align_custodian():
    unaligned_qs = Trade.objects.annotate(
        proper_custodian_id=Subquery(Custodian.objects.filter(mapping__contains=OuterRef("bank")).values("id")[:1])
    ).exclude(custodian__id=F("proper_custodian_id"))

    unaligned_qs.update(custodian__id=F("proper_custodian_id"))


@receiver(post_save, sender="wbportfolio.Claim")
def compute_claimed_shares_on_claim_save(sender, instance, created, raw, **kwargs):
    if not raw and instance.trade:
        instance.trade.save()



import importlib
from contextlib import suppress

from celery import shared_task
from django.db import models
from django.db.models import Exists, OuterRef, Q, QuerySet
from django.dispatch import receiver
from wbfdm.models.instruments.instrument_prices import InstrumentPrice

from wbportfolio.import_export.handlers.fees import FeesImportHandler
from wbportfolio.models.products import Product

from .transactions import Transaction


class ValidFeesQueryset(QuerySet):
    def filter_only_valid_fees(self) -> QuerySet:
        """
        Filter the queryset to remove duplicate in case calculated and non-calculated fees are present for the same date/product/type
        """
        return self.annotate(
            real_fees_exists=Exists(
                self.filter(
                    transaction_subtype=OuterRef("transaction_subtype"),
                    linked_product=OuterRef("linked_product"),
                    fee_date=OuterRef("fee_date"),
                    calculated=False,
                )
            )
        ).filter(Q(calculated=False) | (Q(real_fees_exists=False) & Q(calculated=True)))


class DefaultFeesManager(models.Manager):
    def get_queryset(self) -> ValidFeesQueryset:
        return ValidFeesQueryset(self.model)

    def filter_only_valid_fees(self) -> QuerySet:
        return self.get_queryset().filter_only_valid_fees()


class ValidFeesManager(DefaultFeesManager):
    def get_queryset(self) -> QuerySet:
        return super().get_queryset().filter_only_valid_fees()


class Fees(Transaction):
    import_export_handler_class = FeesImportHandler

    class Type(models.TextChoices):
        TRANSACTION = "TRANSACTION", "Transaction"
        PERFORMANCE_CRYSTALIZED = "PERFORMANCE_CRYSTALIZED", "Performance Crystalized"
        PERFORMANCE = "PERFORMANCE", "Performance"
        MANAGEMENT = "MANAGEMENT", "Management"
        ISSUER = "ISSUER", "Issuer"
        OTHER = "OTHER", "Other"

    transaction_subtype = models.CharField(
        max_length=255, verbose_name="Fees Type", choices=Type.choices, default=Type.MANAGEMENT
    )

    fee_date = models.DateField()  # needed for indexing
    calculated = models.BooleanField(
        default=True,
        help_text="A marker whether the fees were calculated or supplied.",
        verbose_name="Is calculated",
    )

    linked_product = models.ForeignKey(
        "wbportfolio.Product",
        related_name="transactionfees",
        on_delete=models.PROTECT,
        verbose_name="Product",
    )

    class Meta:
        verbose_name = "Fees"
        verbose_name_plural = "Fees"
        indexes = [
            models.Index(fields=["linked_product"]),
            models.Index(fields=["transaction_subtype", "linked_product", "fee_date", "calculated"]),
        ]
        constraints = [
            models.UniqueConstraint(
                fields=["linked_product", "fee_date", "transaction_subtype", "calculated"], name="unique_fees"
            ),
        ]

    objects = DefaultFeesManager()
    valid_objects = ValidFeesManager()

    def save(self, *args, **kwargs):
        self.fee_date = self.transaction_date
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.transaction_date:%d.%m.%Y} - {self.Type[self.transaction_subtype]}: {self.portfolio.name}"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:fees"


class FeeCalculation(models.Model):
    name = models.CharField(max_length=128, verbose_name="Name")
    import_path = models.CharField(max_length=512, verbose_name="Import Path", default="restbench.fees.default")

    @classmethod
    def compute_fee_from_price(cls, price):
        product = Product.objects.get(id=price.instrument.id)
        if (fee_calculation := product.fee_calculation) and (import_path := fee_calculation.import_path):
            calculation_module = importlib.import_module(import_path)
            for new_fees in calculation_module.fees_calculation(price.id):
                Fees.objects.update_or_create(
                    linked_product=new_fees.pop("linked_product"),
                    transaction_date=new_fees.pop("transaction_date"),
                    transaction_subtype=new_fees.pop("transaction_subtype"),
                    calculated=True,
                    defaults=new_fees,
                )

    def __str__(self) -> str:
        return self.name


@shared_task
def compute_fee_from_price_as_task(price_id):
    price = InstrumentPrice.objects.get(id=price_id)
    FeeCalculation.compute_fee_from_price(price)


@receiver(models.signals.post_save, sender="wbfdm.InstrumentPrice")
def update_or_create_fees_post(sender, instance, created, raw, **kwargs):
    """Gets or create the fees for a given price and updates them if necessary"""
    if not raw and created and not instance.calculated and instance.instrument:
        with suppress(Product.DoesNotExist):
            product = Product.objects.get(id=instance.instrument.id)
            if product.fee_calculation:
                compute_fee_from_price_as_task.delay(instance.id)


# @receiver(models.signals.pre_save, sender="wbportfolio.Fees")
# def check_uniqueness(sender, instance, raw, **kwargs):
#     if (
#         Fees.objects.exclude(id=instance.id)
#         .filter(
#             transaction_date=instance.transaction_date,
#             transaction_subtype=instance.transaction_subtype,
#             linked_product=instance.linked_product,
#         )
#         .exists()
#     ):
#         raise ValueError(
#             f"A fees object already exists with date, type and product = {instance.transaction_date}, {instance.transaction_subtype}, {instance.linked_product}"
#         )



from wbportfolio.models import Expiry

from .transactions import (
    TransactionModelSerializer,
    TransactionRepresentationSerializer,
)


class ExpiryRepresentationSerializer(TransactionRepresentationSerializer):
    class Meta:
        model = Expiry
        fields = TransactionRepresentationSerializer.Meta.fields


class ExpiryModelSerializer(TransactionModelSerializer):
    class Meta:
        model = Expiry
        fields = TransactionModelSerializer.Meta.fields



import random
from datetime import timedelta

import factory

from wbportfolio.models import Transaction


class TransactionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Transaction

    currency_fx_rate = 1.0
    portfolio = factory.SubFactory("wbportfolio.factories.PortfolioFactory")
    underlying_instrument = factory.SubFactory("wbfdm.factories.InstrumentFactory")
    transaction_date = factory.Faker("date_object")
    value_date = factory.LazyAttribute(lambda o: o.transaction_date + timedelta(days=1))
    currency = factory.SubFactory("wbcore.contrib.currency.factories.CurrencyFactory")
    total_value = factory.LazyAttribute(lambda o: random.randint(1, 1000))
    total_value_fx_portfolio = factory.LazyAttribute(lambda o: o.currency_fx_rate * o.total_value)
    total_value_gross = factory.LazyAttribute(lambda o: random.randint(1, 1000))
    total_value_gross_fx_portfolio = factory.LazyAttribute(lambda o: o.currency_fx_rate * o.total_value_gross)



from wbcore.metadata.configs.titles import TitleViewConfig


class AssetPositionPandasTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Invested Asset Positions"


class AggregatedAssetPositionLiquidityTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Aggregated Asset Position Liquidity Table"



from wbcore.metadata.configs.titles import TitleViewConfig


class ProductGroupTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return "Product Groups"

    def get_instance_title(self):
        return "Product Group: {{title}} {{umbrella}}"

    def get_create_title(self):
        return "New Product Group"



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class FeesDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="portfolio", label="Portfolio"),
                dp.Field(key="transaction_subtype", label="Type"),
                dp.Field(key="fee_date", label="Date"),
                dp.Field(
                    label="Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="total_value", label="Value"),
                        dp.Field(key="total_value_fx_portfolio", label="Value (Portfolio)"),
                        dp.Field(key="total_value_usd", label="Value ($)"),
                        dp.Field(key="total_value_gross", label="Value Gross"),
                        dp.Field(key="total_value_gross_fx_portfolio", label="Value Gross (Portfolio)"),
                        dp.Field(key="total_value_gross_usd", label="Value Gross ($)"),
                    ],
                ),
                dp.Field(
                    label="Information",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="currency_fx_rate", label="FX rate"),
                        dp.Field(key="currency", label="Currency"),
                        dp.Field(key="calculated", label="Calculated"),
                    ],
                ),
            ],
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(2, "portfolio"), "transaction_subtype"],
                ["fee_date", "value_date", "book_date"],
                ["currency_fx_rate", "currency", "."],
                ["total_value", "total_value_fx_portfolio", "."],
                ["total_value_gross", "total_value_gross_fx_portfolio", "."],
                [repeat_field(3, "comment")],
            ]
        )


class FeesPortfolioDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="transaction_subtype", label="Type"),
                dp.Field(key="fee_date", label="Date"),
                dp.Field(
                    label="Value",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(
                            label="Net",
                            open_by_default=False,
                            key=None,
                            children=[
                                dp.Field(key="total_value", label="Value"),
                                dp.Field(key="total_value_fx_portfolio", label="Value (Portfolio)", show="open"),
                                dp.Field(key="total_value_usd", label="Value ($)", show="open"),
                            ],
                        ),
                        dp.Field(
                            label="Gross",
                            open_by_default=False,
                            key=None,
                            children=[
                                dp.Field(key="total_value_gross", label="Value Gross", show="open"),
                                dp.Field(
                                    key="total_value_gross_fx_portfolio", label="Value Gross (Portfolio)", show="open"
                                ),
                                dp.Field(key="total_value_gross_usd", label="Value Gross ($)", show="open"),
                            ],
                        ),
                    ],
                ),
                dp.Field(
                    label="Information",
                    open_by_default=False,
                    key=None,
                    children=[
                        dp.Field(key="calculated", label="Calculated"),
                        dp.Field(key="currency_fx_rate", label="FX rate", show="open"),
                        dp.Field(key="currency", label="Currency", show="open"),
                        dp.Field(key="linked_product", label="Product", show="open"),
                    ],
                ),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                [repeat_field(3, "transaction_subtype")],
                ["fee_date", "value_date", "book_date"],
                ["currency_fx_rate", "currency", "."],
                ["total_value", "total_value_fx_portfolio", "."],
                ["total_value_gross", "total_value_gross_fx_portfolio", "."],
                [repeat_field(3, "comment")],
            ]
        )


class FeesAggregatedPortfolioPandasDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="fee_date", label="Date"),
                dp.Field(key="TRANSACTION", label="Transactions"),
                dp.Field(key="PERFORMANCE_CRYSTALIZED", label="Performance"),
                dp.Field(key="PERFORMANCE", label="Performance Crystalized"),
                dp.Field(key="MANAGEMENT", label="Management"),
                dp.Field(key="ISSUER", label="Issuer"),
                dp.Field(key="OTHER", label="Other"),
                dp.Field(key="total", label="Total"),
            ]
        )



from datetime import timedelta

import pytest
from faker import Faker

from wbportfolio.factories import ProductFactory

fake = Faker()


class PortfolioTestMixin:
    @pytest.fixture
    def active_product(self, weekday):
        return ProductFactory.create(inception_date=weekday - timedelta(days=30), delisted_date=None)



from datetime import date, timedelta

from psycopg.types.range import DateRange
from wbcore import filters as wb_filters

from wbportfolio.models import Fees, Portfolio, Trade, Transaction


def get_transaction_gte_default(field, request, view):
    filter_date = date.today() - timedelta(days=90)
    qs = Transaction.objects.none()
    if "instrument_id" in view.kwargs:
        qs = Transaction.objects.filter(underlying_instrument__id=view.kwargs["instrument_id"])
    elif "portfolio_id" in view.kwargs:
        qs = Transaction.objects.filter(portfolio__id=view.kwargs["portfolio_id"])
    if qs.exists():
        filter_date = qs.earliest("transaction_date").transaction_date
    return filter_date


def get_transaction_underlying_type_choices(*args):
    models = [Fees, Trade]
    choices = []
    for model in models:
        for choice in model.Type.choices:
            choices.append(choice)
    return choices


def get_transaction_lte_default(field, request, view):
    filter_date = date.today() + timedelta(days=7)
    qs = Transaction.objects.none()
    if "instrument_id" in view.kwargs:
        qs = Transaction.objects.filter(underlying_instrument__id=view.kwargs["instrument_id"])
    elif "portfolio_id" in view.kwargs:
        qs = Transaction.objects.filter(portfolio__id=view.kwargs["portfolio_id"])
    if qs.exists():
        filter_date = qs.latest("transaction_date").transaction_date
    return filter_date


def get_transaction_default_date_range(*args, **kwargs):
    return DateRange(get_transaction_gte_default(*args, **kwargs), get_transaction_lte_default(*args, **kwargs))


class TransactionFilterSet(wb_filters.FilterSet):
    transaction_date = wb_filters.DateRangeFilter(
        method=wb_filters.DateRangeFilter.base_date_range_filter_method,
        label="Date Range",
        initial=get_transaction_default_date_range,
    )

    portfolio = wb_filters.ModelChoiceFilter(
        label="Portfolio",
        queryset=Portfolio.objects.all(),
        endpoint=Portfolio.get_representation_endpoint(),
        value_key=Portfolio.get_representation_value_key(),
        label_key=Portfolio.get_representation_label_key(),
    )
    total_value_usd__gte = wb_filters.NumberFilter(
        lookup_expr="gte", field_name="total_value_usd", label="Total Value ($)"
    )
    total_value_usd__lte = wb_filters.NumberFilter(
        lookup_expr="lte",
        label="Total Value ($)",
        field_name="total_value_usd",
    )

    transaction_underlying_type = wb_filters.ChoiceFilter(
        label="Underlying Type",
        choices=get_transaction_underlying_type_choices(),
        method="filter_transaction_underlying_type",
    )

    def filter_transaction_underlying_type(self, queryset, name, value):
        if value:
            queryset = queryset.filter(transaction_underlying_type=value)
        return queryset

    class Meta:
        model = Transaction
        fields = {
            "transaction_type": ["exact"],
            "underlying_instrument": ["exact"],
            "currency": ["exact"],
            "total_value": ["gte", "exact", "lte"],
            "total_value_fx_portfolio": ["gte", "exact", "lte"],
        }


class TransactionPortfolioFilterSet(TransactionFilterSet):
    portfolio = transaction_date__gte = transaction_date__lte = None
    transaction_date = wb_filters.DateFilter(
        label="Transaction Date",
        lookup_expr="exact",
        field_name="transaction_date",
        initial=get_transaction_lte_default,
        required=True,
    )



from typing import Optional

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
)
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class RegisterDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="register_reference", label="Register Reference"),
                dp.Field(key="register_name_1", label="Register Name 1"),
                dp.Field(key="register_name_2", label="Register Name 2"),
                dp.Field(key="custodian_reference", label="Custodian Reference"),
                dp.Field(key="custodian_name_1", label="Custodian Name 1"),
                dp.Field(key="custodian_name_2", label="Custodian Name 2"),
                dp.Field(key="outlet_reference", label="Outlet Reference"),
                dp.Field(key="outlet_name", label="Outlet Name"),
                dp.Field(key="status", label="Status"),
                dp.Field(key="status_message", label="Status Message"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display(
            [
                ["register_reference", "register_name_1", "register_name_2"],
                ["global_register_reference", "external_register_reference", "."],
                ["citizenship", "residence", "investor_type"],
                [repeat_field(3, "custodian_section")],
                [repeat_field(3, "outlet_section")],
                [repeat_field(3, "information_section")],
            ],
            [
                create_simple_section(
                    "custodian_section",
                    "Custodian",
                    [
                        ["custodian_reference", "custodian_name_1", "custodian_name_2"],
                        [repeat_field(3, "custodian_address")],
                        ["custodian_postcode", "custodian_city", "custodian_country"],
                    ],
                ),
                create_simple_section(
                    "outlet_section",
                    "Outlet",
                    [
                        ["sales_reference", repeat_field(2, "dealer_reference")],
                        ["outlet_reference", repeat_field(2, "outlet_name")],
                        [repeat_field(3, "outlet_address")],
                        ["outlet_postcode", "outlet_city", "outlet_country"],
                    ],
                ),
                create_simple_section(
                    "information_section",
                    "Information",
                    [
                        ["opened", repeat_field(2, "status")],
                        [".", "opened_reference_1", "opened_reference_2"],
                        [".", "updated_reference_1", "updated_reference_2"],
                        [repeat_field(3, "status_message")],
                    ],
                ),
            ],
        )



import re
from datetime import datetime

import pandas as pd

from wbportfolio.models import FeeProductPercentage, Fees, Product

FIELD_MAP = {
    "Transaction_Date": "transaction_date",
    "Trade_Date": "book_date",
    "Value_Date": "value_date",
    "Quantity": "total_value",
    "Currency": "currency__key",
    "Portfolio_Identifier": "linked_product",
    "Booking_Comment": "comment",
}


def parse(import_source):
    df = pd.read_csv(import_source.file, encoding="utf-16", delimiter=";")
    df = df.loc[(df["Order_Type"] == "xfermon_client") & (df["Bookkind"] == "forex_impl_macc"), :]
    date_range_regex = import_source.source.import_parameters.get(
        "date_range_regex", r"\((\b\d{2}\.\d{2}\.\d{4}\b) to (\b\d{2}\.\d{2}\.\d{4}\b)\)"
    )
    data = []
    if not df.empty:
        df = df.rename(columns=FIELD_MAP)
        df.linked_product = df.linked_product.apply(lambda x: Product.objects.filter(identifier=x).first())
        df["transaction_date"] = pd.to_datetime(df["transaction_date"], dayfirst=True)
        df["book_date"] = pd.to_datetime(df["book_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["value_date"] = pd.to_datetime(df["value_date"], dayfirst=True).dt.strftime("%Y-%m-%d")
        df["portfolio"] = df.linked_product.apply(lambda x: x.primary_portfolio if x else None)
        df["base_management_fees"] = df.apply(
            lambda row: float(
                row["linked_product"].get_fees_percent(row["transaction_date"], FeeProductPercentage.Type.MANAGEMENT)
            ),
            axis=1,
        )
        df["base_bank_fees"] = df.apply(
            lambda row: float(
                row["linked_product"].get_fees_percent(row["transaction_date"], FeeProductPercentage.Type.BANK)
            ),
            axis=1,
        )
        df.total_value = -df.total_value

        df_management = df.copy()
        df_management["transaction_subtype"] = Fees.Type.MANAGEMENT
        df_management["total_value"] = (
            df_management["total_value"]
            * df_management["base_management_fees"]
            / (df_management["base_management_fees"] + df_management["base_bank_fees"])
        )

        df_bank = df.copy()
        df_bank["transaction_subtype"] = Fees.Type.ISSUER
        df_bank["total_value"] = (
            df_bank["total_value"]
            * df_bank["base_bank_fees"]
            / (df_bank["base_management_fees"] + df_bank["base_bank_fees"])
        )

        df = pd.concat([df_management, df_bank], axis=0)
        df = df.drop(columns=df.columns.difference(["transaction_subtype", *FIELD_MAP.values()]))

        df["linked_product"] = df.linked_product.apply(lambda x: x.id)
        df = df.dropna(subset=["linked_product"])

        for row in df.to_dict("records"):
            # The fee are sometime aggregated. The aggregate date range information is given in the comment. We therefore try to extract it and duplicate the averaged fees
            if match := re.search(date_range_regex, row["comment"]):
                from_date = datetime.strptime(match.group(1), "%d.%m.%Y").date()
                to_date = datetime.strptime(match.group(2), "%d.%m.%Y").date()
            else:
                from_date = (row["transaction_date"] - pd.tseries.offsets.BDay(1)).date()
                to_date = row["transaction_date"]

            dates = pd.date_range(from_date, to_date, freq="B", inclusive="right")
            for ts in pd.date_range(from_date, to_date, freq="B", inclusive="right"):
                row_copy = row.copy()
                row_copy["transaction_date"] = ts.strftime("%Y-%m-%d")
                row_copy["total_value"] = row["total_value"] / len(dates)
                data.append(row_copy)

    return {"data": data}
